/*
 * exrtoiff.cpp
 *
 * $Id: exrtotiff.cpp,v 1.1 2005/04/30 18:27:53 kegan Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <tiffio.h>
#include <ImfInputFile.h>
#include <ImfChannelList.h>
#include <ImfFrameBuffer.h>
#include <half.h>
#include <ImathMath.h>
#include <ImathFun.h>
#include <halfFunction.h>

#include "pbrt.h"
#include "film.h"
#include "paramset.h"

using namespace Imf;
using namespace Imath;

static bool ReadEXR(const char *name, float *&rgba, int &xRes, int &yRes,
    bool &hasAlpha);
static void WriteTIFF(const char *name, float *rgba, int xRes, int yRes,
    bool hasAlpha, bool useAlpha);
static void ApplyExposurePipeline(float* rgb, int xRes, int yRes,
    float exposure, float gamma, bool anerGamma, float linearScale);

static void usage() {
  fprintf( stderr, "usage: exrtotiff [options] <input.exr> <output.tiff>\n" );
  fprintf( stderr, "Supported options:\n");
  fprintf( stderr, "\t-scale scale\n" );
  fprintf( stderr, "\t-gamma gamma\n" );
  fprintf( stderr, "\t-tonemap mapname\n" );
  fprintf( stderr, "\t-param name value (set tone map parameter)\n" );
  fprintf( stderr, "\t-bloomRadius radius\n" );
  fprintf( stderr, "\t-bloomWeight weight\n" );
  fprintf( stderr, "\t-bg bg-grey color\n");
  fprintf( stderr, "\t-useAlpha [0.0 | 1.0]\n");
  fprintf( stderr, "\t-exposure filmExposure\n");
  fprintf( stderr, "\t-weightChannels [0.0 | 1.0]\n");
  fprintf( stderr, "\t-anerGamma [0.0 | 1.0]\n");
  fprintf( stderr, "\t-linear scale\n");
  exit(1);
}

int main(int argc, char *argv[]) 
{
    float scale = 1.f, gamma = 2.2f;
    float bloomRadius = 0.f, bloomWeight = .2f;
    float bggray = -1.f;
    char *toneMap = NULL;
    float useAlpha = 1.f;  // 0.0 = no alpha, 1.0 = use alpha
    float anerGamma = 0.f;  // 0.0 = correct, 1.0 = aner
    float linearScale = 0.f; // if scale > 0 then use linear scale pipeline
    float exposure = -100.f; // if exposure > -100 then use exrdisplay pipeline
    ParamSet toneMapParams;

    int argNum = 1;
    while (argNum < argc && argv[argNum][0] == '-') {
#define ARG(name, var) \
      else if (!strcmp(argv[argNum], "-" name)) { \
	if (argNum+1 == argc) \
	  usage(); \
	var = atof(argv[argNum+1]); \
	++argNum; \
      }

	if (!strcmp(argv[argNum], "-tonemap")) {
	    if (argNum+1 == argc)
		usage();
	    toneMap = argv[argNum+1];
	    ++argNum;
	}
	else if (!strcmp(argv[argNum], "-param")) {
	    if (argNum+2 >= argc)
		usage();
	    float val = atof(argv[argNum+2]);
	    toneMapParams.AddFloat(argv[argNum+1], &val);
	    argNum += 2;
	}
	ARG("scale", scale)
	ARG("gamma", gamma)
	ARG("bg", bggray)
	ARG("bloomRadius", bloomRadius)
	ARG("bloomWeight", bloomWeight)
	ARG("useAlpha", useAlpha)
	ARG("exposure", exposure)
	ARG("anerGamma", anerGamma)
	ARG("linear", linearScale)
	else 
	    usage();
      ++argNum;

    }
    if (argNum + 2 > argc) usage();

    char *inFile = argv[argNum], *outFile = argv[argNum+1]; 	  
    float *rgba;
    int xRes, yRes;
    bool hasAlpha;
    pbrtInit();

    if (ReadEXR(inFile, rgba, xRes, yRes, hasAlpha)) {
	float *rgb = new float[xRes*yRes*3];
	for (int i = 0; i < xRes*yRes; ++i) {
	    for (int j = 0; j < 3; ++j) {
		rgb[3*i+j] = scale * rgba[4*i+j];
//CO		if (rgba[4*i+3] != 0.f)
//CO		    rgb[3*i+j] /= rgba[4*i+3];
		if (bggray > 0)
		    rgb[3*i+j] = rgba[4*i+3] * rgb[3*i+j] + (1.f - rgba[4*i+3]) * bggray;
	    }
	    if (bggray > 0)
		rgba[4*i+3] = 1.f;
	}

        if (exposure > -100.f || linearScale > 0.0f) {
            // taken from exrdisplay
            ApplyExposurePipeline(rgb, xRes, yRes, exposure, gamma,
                (anerGamma != 0.0), linearScale);
        } else {
            ApplyImagingPipeline(rgb, xRes, yRes, NULL, bloomRadius,
                                 bloomWeight, toneMap, &toneMapParams,
                                 gamma, 0.f, 255);
        }

	for (int i = 0; i < xRes*yRes; ++i) {
	    for (int j = 0; j < 3; ++j) {
		rgba[4*i+j] = rgb[3*i+j];
                // XXX TODO this can cause values greater than 255.0
		//if (rgba[4*i+3] != 0.f)
		//    rgba[4*i+j] /= rgba[4*i+3];
	    }
	    // rgba[4*i+3] *= 255.f;

            // if useAlpha is 0.0 set alpha channel to 255
            rgba[4*i+3] = (useAlpha == 0.0) ? (255.f) : (rgba[4*i+3] * 255.f);
	}

        // writing 3 channels doesn't seem to work, so set useAlpha to 1.0
	WriteTIFF(outFile, rgba, xRes, yRes, hasAlpha, 1.0);
    }
    pbrtCleanup();
    return 0;
}

void WriteTIFF(const char *name, float *rgba, int XRes, int YRes,
    bool hasAlpha, bool useAlpha)
{
    // Open 8-bit TIFF file for writing
    TIFF *tiff = TIFFOpen(name, "w");
    if (!tiff) {
	fprintf(stderr, "Unable to open TIFF %s for writing", name);
	return;
    }

    int alphaSkip = (hasAlpha && !useAlpha) ? 1 : 0;
    int nChannels = (hasAlpha && useAlpha) ? 4 : 3;

    TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, nChannels);
    if (hasAlpha) {
	short int extra[] = { EXTRASAMPLE_ASSOCALPHA };
	TIFFSetField(tiff, TIFFTAG_EXTRASAMPLES, (short)1, extra);
    }
    // Write image resolution information
    TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, XRes);
    TIFFSetField(tiff, TIFFTAG_IMAGELENGTH, YRes);
    TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    // Set Generic TIFF Fields
    TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, 1);
    TIFFSetField(tiff, TIFFTAG_XRESOLUTION, 1.f);
    TIFFSetField(tiff, TIFFTAG_YRESOLUTION, 1.f);
    TIFFSetField(tiff, TIFFTAG_RESOLUTIONUNIT, (short)1);
    TIFFSetField(tiff, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(tiff, TIFFTAG_ORIENTATION, (int)ORIENTATION_TOPLEFT);
    // Write 8-bit scanlines
    u_char *buf = new u_char[nChannels * XRes];
    for (int y = 0; y < YRes; ++y) {
	u_char *bufp = buf;
	for (int x = 0; x < XRes; ++x) {
	    // Pack 8-bit pixels samples into buf
	    for (int s = 0; s < nChannels; ++s)
		*bufp++ = (u_char)*rgba++;
            rgba += alphaSkip;
	}
	TIFFWriteScanline(tiff, buf, y, 1);
    }
    // Close 8-bit TIFF file
    delete[] buf;
    TIFFClose(tiff);
}

static bool ReadEXR(const char *name, float *&rgba, int &xRes, int &yRes, bool &hasAlpha)
{
    InputFile file(name);
    Box2i dw = file.header().dataWindow();
    xRes = dw.max.x - dw.min.x + 1;
    yRes = dw.max.y - dw.min.y + 1;

    rgba = new float[4 * xRes * yRes];

    // for now...
    hasAlpha = true;
    int nChannels = 4;

    float *hp = rgba - nChannels * (dw.min.x + dw.min.y * xRes);

    FrameBuffer frameBuffer;
    frameBuffer.insert("R", Slice(FLOAT, (char *)hp,
				  4*sizeof(float), xRes * 4 * sizeof(float), 1, 1, 0.0));
    frameBuffer.insert("G", Slice(FLOAT, (char *)hp+sizeof(float),
				  4*sizeof(float), xRes * 4 * sizeof(float), 1, 1, 0.0));
    frameBuffer.insert("B", Slice(FLOAT, (char *)hp+2*sizeof(float),
				  4*sizeof(float), xRes * 4 * sizeof(float), 1, 1, 0.0));
    frameBuffer.insert("A", Slice(FLOAT, (char *)hp+3*sizeof(float),
				  4*sizeof(float), xRes * 4 * sizeof(float), 1, 1, 1.0));

    file.setFrameBuffer(frameBuffer);
    file.readPixels(dw.min.y, dw.max.y);

    return rgba;
}


// taken from ImageView.cpp in the exrdisplay directory in the OpenEXR package
static float knee(double x, double f)
{
    return float (Imath::Math<double>::log (x * f + 1) / f);
}

// taken from ImageView.cpp in the exrdisplay directory in the OpenEXR package
static float findKneeF(float x, float y)
{
    float f0 = 0;
    float f1 = 1;

    while (knee (x, f1) > y)
    {
	f0 = f1;
	f1 = f1 * 2;
    }

    for (int i = 0; i < 30; ++i)
    {
	float f2 = (f0 + f1) / 2;
	float y2 = knee (x, f2);

	if (y2 < y)
	    f1 = f2;
	else
	    f0 = f2;
    }

    return (f0 + f1) / 2;
}


// taken from comments and code in ImageView.cpp in the
// exrdisplay directory in the OpenEXR package
static void ApplyExposurePipeline(float* rgb, int xRes, int yRes,
    float exposure, float gamma, bool anerGamma, float linearScale)
{

    if (linearScale > 0.0f) {
        fprintf(stderr, "linearScale %.3f\n", linearScale);
        linearScale *= 255.0f;
        int numVals= xRes * yRes;
        for (int i = 0; i < numVals; i++) {
            rgb[i * 3 + 0] = clamp(rgb[i * 3 + 0] * linearScale, 0.f, 255.f);
            rgb[i * 3 + 1] = clamp(rgb[i * 3 + 1] * linearScale, 0.f, 255.f);
            rgb[i * 3 + 2] = clamp(rgb[i * 3 + 2] * linearScale, 0.f, 255.f);
        }

        return;
    }


    float defog = 0.0;  // there is also a per-channel fog multiplier in OpenEXR
    float kneeLow = 0.0;
    float kneeHigh = 5.0;
    float gammaInv = 1.0 / gamma;

    float exposureMult = Imath::Math<float>::pow(2.f, exposure + 2.47393f);
    float kl = Imath::Math<float>::pow(2.f, kneeLow);
    float f = findKneeF(Imath::Math<float>::pow(2.f, kneeHigh) - kl,
                        Imath::Math<float>::pow(2.f, 3.5f) - kl);

    if (anerGamma) {
        int numVals= xRes * yRes;
        for (int i = 0; i < numVals; i++) {
            /*
            float y = (2.0f/9.0f)*original[0] + (1.0f/9.0f)*original[1] +
            (6.0f/9.0f)*original[2];
            // float y = 0.3f*original[0] + 0.6f*original[1] + 0.1f*original[2];
            if (y <= 0) {
            original = vec3ZERO;
            } else {
            y *= 0.33333333f;
            float x = y;
            x *= exposureMult;
            x = powf(x, gammaInv);
            x *= 84.66f / 255.0f;
            // scale and clamp
            CLAMP(0.f, x, 1.0f);
            mult(original, x / y);
            */

            //float y = (rgb[i * 3 + 0] * 2.0 +
            //        rgb[i * 3 + 1] * 1.0 +
            //        rgb[i * 3 + 2] * 6.0) / 9.0;
            float y = (rgb[i * 3 + 0] + rgb[i * 3 + 1] + rgb[i * 3 + 2]) / 3.0;
            if (y <= 0) {
                rgb[i * 3 + 0] = 0.0;
                rgb[i * 3 + 1] = 0.0;
                rgb[i * 3 + 2] = 0.0;
            } else {
                //y *= 0.33333333f;

                float x = y;

                // exposure
                x *= exposureMult;

                // gamma
                x = powf(x, gammaInv);
                x *= 84.66f;

                // scale and clamp
                rgb[i * 3 + 0] *= x / y;
                rgb[i * 3 + 1] *= x / y;
                rgb[i * 3 + 2] *= x / y;
                rgb[i * 3 + 0] = clamp(rgb[i * 3 + 0], 0.f, 255.f);
                rgb[i * 3 + 1] = clamp(rgb[i * 3 + 1], 0.f, 255.f);
                rgb[i * 3 + 2] = clamp(rgb[i * 3 + 2], 0.f, 255.f);
            }
        }
    } else {
        int numVals= xRes * yRes* 3;
        for (int i = 0; i < numVals; i++) {
            // defog
            float x = max(0.f, (rgb[i] - defog));

            // exposure
            x *= exposureMult;

            // knee
            if (x > kl) {
                x = kl + knee(x - kl, f);
            }

            // gamma
            x = powf(x, gammaInv);

            // scale and clamp
            rgb[i] = clamp(x * 84.66f, 0.f, 255.f);
        }
    }
}


