// Copyright (c) 2009, Yu-Ting Tseng and Kevin Egan
// MIT License, see LICENSE file for details


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>

#include <algorithm>
#include <vector>
#include <iostream>
#include <fstream>
#include <tclap/CmdLine.h>

#include "util.h"
#include "const.h"
#include "render-region.h"
#include "call_pixie.h"
#include "config.h"

using namespace std;

const bool OUTPUT_FILTER_SIZE_MAP = false;
const bool OUTPUT_SAMPLE_DENSITY_MAP = false;

const bool SHEAR_SAMPLING = true;
const bool STRATIFIED_SAMPLING = false;
const bool ADAPTIVE_SAMPLING = true;
const bool SEGMENTATION = false;
const bool RERUN_FIRST_PASS = false;
const bool RERUN_SECOND_PASS = false;

TCLAP::CmdLine *cmd = NULL;
TCLAP::CmdLine &CmdLineSingleton() {
  if (!cmd)
    cmd = new TCLAP::CmdLine("PRT Shader", ' ', "0.1");
  return *cmd;
}

TCLAP::SwitchArg norerun_pixie_pass1_arg(
    "","norerun1","Do not re-run the first pass sampling", CmdLineSingleton(), false);
TCLAP::SwitchArg norerun_pixie_pass2_arg(
    "","norerun2","Do not re-run the second pass sampling", CmdLineSingleton(), false);
TCLAP::SwitchArg output_sampling_density_map_arg(
    "","output_sampling_density","output the sampling density map", CmdLineSingleton(), false);
TCLAP::SwitchArg output_filter_width_map_arg(
    "","output_filter_width","output the filter width at each pixel", CmdLineSingleton(), false);
TCLAP::SwitchArg use_sinc_filter_arg(
    "","sinc","Whether to use Sinc function as the filter.", CmdLineSingleton(), false);
TCLAP::SwitchArg disable_blur_arg(
    "","noblur","Whether to disable a_min and a_max blurring.", CmdLineSingleton(), false);
TCLAP::SwitchArg norefine_filter_width_arg(
    "","norfw","Whether to not refine filter widths.", CmdLineSingleton(), false);
TCLAP::SwitchArg nogreedy_refine_filter_arg(
    "","nogreedy","Whether to not to use greedy algorithm to refine filter widths.", CmdLineSingleton(), false);
TCLAP::ValueArg<int> render_region_arg(
    "r","renderregion", "Render region index", false, 2, "int", CmdLineSingleton());
TCLAP::ValueArg<string> config_filename_arg(
    "c","config", "Config filename", false, "quad.config", "string", CmdLineSingleton());
TCLAP::ValueArg<double> temporal_sampling_rate_factor_arg(
    "t","tf", "Temporal sampling rate factor", false, 1, "double", CmdLineSingleton());
TCLAP::ValueArg<double> spatial_sampling_rate_factor_arg(
    "s","sf", "Spatial sampling rate factor", false, 1, "double", CmdLineSingleton());
TCLAP::ValueArg<string> manual_render_region_arg(
    "","mrr", "Manual Render Region", false, "", "string", CmdLineSingleton());
TCLAP::ValueArg<string> output_filename_arg(
    "o","output", "Output filename", false, "final.exr", "string", CmdLineSingleton());
TCLAP::SwitchArg keep_sample_files_arg(
    "","keepfiles","Whether to keep temp sample files.", CmdLineSingleton(), false);
TCLAP::SwitchArg uniform_arg(
    "","uniform","Whether to use uniform sampling (default is halton).", CmdLineSingleton(), false);
TCLAP::SwitchArg nobucket_arg(
    "","nobucket","Whether to turn off bucketing inside the filter.", CmdLineSingleton(), false);
TCLAP::SwitchArg oldbucket_arg(
    "","oldbucket","Whether to use old bucket filter reweighting scheme.", CmdLineSingleton(), false);
TCLAP::ValueArg<double> animation_start_arg(
    "","animationstart", "Start (relative to .rib frames) of animation", false, -1, "double", CmdLineSingleton());
TCLAP::ValueArg<double> animation_length_arg(
    "","animationlength", "Frames of animation for this image (relative to .rib frames)", false, -1, "double", CmdLineSingleton());

TCLAP::ValueArg<int> max_samples_arg(
    "","maxsamples", "Maximum number of samples per pixel", false, 32, "int", CmdLineSingleton());
TCLAP::ValueArg<int> set_samples_arg(
    "","setsamples", "Set samples per pixel, must use --halton", false, -1, "int", CmdLineSingleton());
TCLAP::ValueArg<int> stratified_x_arg(
    "","stratifiedx", "Set stratified spatial density per pixel", false, -1, "int", CmdLineSingleton());
TCLAP::ValueArg<int> stratified_t_arg(
    "","stratifiedt", "Set stratified time density per pixel", false, -1, "int", CmdLineSingleton());
TCLAP::ValueArg<int> initial_samples_arg(
    "","initialsamples", "Set samples per pixel for initial sampling", false, 2, "int", CmdLineSingleton());

// this can be necessary if Pixie doesn't calculate the motion correctly
// so you use PRMan for initial sampling (such as the teapot scene)
TCLAP::ValueArg<string> hack_initial_arg(
    "","hackinitial", "Read in initial sample data from file", false, "", "string", CmdLineSingleton());

float GetOffset() { return STRATIFIED_SAMPLING ? fmin(0.99999, drand48()) : 0.5; }
void ConstructSegments(vector<FreqSample*>* samples_in_grid,
                       vector<TimeSegmentAtPixel*>* segments);

vector<TimeSegmentAtPixel*>* segments;

// Sample comparator that sorts samples in the same voxel based on time.
bool SampleCMP(const FreqSample *a, const FreqSample *b) {
  return a->shift_time < b->shift_time;
}

timeval starttime;

float initwalltime() {
    gettimeofday(& starttime, NULL);
}

float getwalltime() {
    timeval buf;
    gettimeofday(& buf, NULL);
    return (buf.tv_sec - starttime.tv_sec) + (buf.tv_usec - starttime.tv_usec) * 1e-6;
}


int main(int argc, char* argv[]) {
  initwalltime();
  float program_start = getwalltime();

  CmdLineSingleton().parse( argc, argv );
  LoadConfig(config_filename_arg.getValue().c_str());


  // Specify which region to render. Useful for debugging.
  setRenderRegion(render_region_arg.getValue());
  setRenderRegionString(manual_render_region_arg.getValue().c_str());

  // Allocating spaces
  vector<InputSample> inSamples;
  vector<FreqSample> outSamples;
  float* img = new float[IMAGE_WIDTH * IMAGE_HEIGHT * 4];
  vector<FreqSample*>* samples_in_grid;
  samples_in_grid = new vector<FreqSample*>[IMAGE_WIDTH * IMAGE_HEIGHT];
  segments = new vector<TimeSegmentAtPixel*>[IMAGE_HEIGHT * IMAGE_WIDTH];
  
  // Specify sinc as the filter function. Note that OMEGA_T_MAX needs to be
  // updated if we change the filter function.
  FilterFunc filter_func =
      use_sinc_filter_arg.getValue() ?
      SincWindow : GaussianWindow;


  // Generate 1st pass samples (1/2 res in width and height).
  GenerateSamples(IMAGE_WIDTH, IMAGE_HEIGHT, 0, 1,
                  IMAGE_WIDTH,
                  IMAGE_HEIGHT,
                  initial_samples_arg.getValue(),
                  RENDER_LEFT, RENDER_RIGHT,
                  RENDER_TOP, RENDER_BOTTOM,
                  STRATIFIED_SAMPLING, inSamples);

  // Render 1st pass samples.
  float initial_sample_start = getwalltime();
  if (hack_initial_arg.getValue() != "") {
    HackFreqSamples(outSamples, hack_initial_arg.getValue().c_str(),
        initial_samples_arg.getValue());
  } else if (animation_start_arg.getValue() > 0 && animation_length_arg.getValue() > 0) {
      RenderFreqSamplesMultiFile(inSamples, outSamples,
          IMAGE_WIDTH, IMAGE_HEIGHT, RIB_FILENAME,
          PASS1_SAMPLE_INPUT_FILENAME, PASS1_SAMPLE_RESULTS_FILENAME,
          !norerun_pixie_pass1_arg.getValue(), keep_sample_files_arg.getValue(),
          animation_start_arg.getValue(), 1.0 / animation_length_arg.getValue());
  } else {
      RenderFreqSamples(inSamples, outSamples, IMAGE_WIDTH, IMAGE_HEIGHT,
          RIB_FILENAME,
          PASS1_SAMPLE_INPUT_FILENAME, PASS1_SAMPLE_RESULTS_FILENAME,
          !norerun_pixie_pass1_arg.getValue(), keep_sample_files_arg.getValue());
  }
  float initial_sample_end = getwalltime();
  float initial_sample_sec = (initial_sample_end - initial_sample_start);
  fprintf(stderr, "construct segments\n");

  // Index samples in grid.
  for (int i = 0; i < outSamples.size(); i++) {
    FreqSample curSample = outSamples[i];
    int xpix = (int) (curSample.x);
    int ypix = (int) (curSample.y);
    samples_in_grid[ypix * IMAGE_WIDTH + xpix].push_back(&outSamples[i]);
  }

  // Put samples in segments and compute a_min, a_max and avg_motion.
  ConstructSegments(samples_in_grid, segments);

  fprintf(stderr, "blur amin/amax/omega\n");
  // blur a_min and a_max in a conservative way
  int blurRadius = 5;
  if (!disable_blur_arg.getValue())
    BlurMotion(segments, blurRadius);

  fprintf(stderr, "refine filter widths\n");
    float refine_start = getwalltime();
  // Refine filter widths.
  for (int y = 0; y < IMAGE_HEIGHT; y++) {
    for (int x = 0; x < IMAGE_WIDTH; x++) {
      if (!IsInRenderBound(x, y)) continue;
      if (segments[y * IMAGE_WIDTH + x].size() != 1) {
        cout << "ERROR: more than one segment." << endl;
        exit(1);
      }
      TimeSegmentAtPixel *segment = segments[y * IMAGE_WIDTH + x][0];
      float a_min = segment->a_min;
      float a_max = segment->a_max;
      // Get the widest filter width predicted by the theorem.
      segment->wide_filter_radius = GetOptimalFilterWidth(a_min, a_max, segment->omega_x_max);
      segment->refined_filter_radius = segment->wide_filter_radius;

      // Trim the filter width to cover the region between TIME_START and
      // TIME_END only.  We only want to trim refined_filter_radius,
      // not wide_filter_radius.
      segment->refined_filter_radius = fmin(GetFilterShearedSlope(a_min, a_max) * (TIME_END - TIME_START) / MOTION_DIRECTION_LOBES,
                 segment->refined_filter_radius);

      segment->filter_calculated = false;
      if (ShouldUseAxisAlignedFilter(*segment)) {
          // nothing more to do for this guy
          segment->filter_marked = true;
      } else {
          segment->filter_marked = false;
      }
    }
  }

    float maxWidth = (stratified_x_arg.getValue() > 0) ? 10 : 1e10;
    if (!norefine_filter_width_arg.getValue()) {
        for (int y = 0; y < IMAGE_HEIGHT; y++) {
            for (int x = 0; x < IMAGE_WIDTH; x++) {
                if (nogreedy_refine_filter_arg.getValue()) {
                    if (y % 50 == 0 && x == 0) {
                        cout << "Slow refine filter width. y = " << y << endl;
                    }
                    RefineFilterWidth(segments[y * IMAGE_WIDTH + x][0], samples_in_grid, maxWidth);
                } else {
                    if (y % 50 == 0 && x == 0) {
                        cout << "Greedy refine filter width. y = " << y << endl;
                    }
                    GreedyRefineFilterWidth(segments, x, y, samples_in_grid, maxWidth);
                }
            }
        }
        float* img2 = new float[IMAGE_WIDTH * IMAGE_HEIGHT * 4];
        memset(img2, 0, IMAGE_WIDTH * IMAGE_HEIGHT * 4 * sizeof(float));

        for (int py = 0; py < IMAGE_HEIGHT; py++) {
          for (int px = 0; px < IMAGE_WIDTH; px++) {
            float *pixel_color = img2 + (py * IMAGE_WIDTH + px) * 4;
            TimeSegmentAtPixel* cursegment = segments[py * IMAGE_WIDTH + px][0];
            pixel_color[0] = cursegment->filter_calculated;
            pixel_color[1] = cursegment->filter_marked;
          }
        }
        WriteExr("./greedy.exr", img2, IMAGE_WIDTH, IMAGE_HEIGHT);
    }
    float refine_end = getwalltime();
    float refine_sec = (refine_end - refine_start);


  fprintf(stderr, "clear image\n");
  ClearImage(img, IMAGE_WIDTH, IMAGE_HEIGHT);
  delete[] samples_in_grid;
  outSamples.clear();

  float* num_temporal_samples_required_in_grid = new float[IMAGE_WIDTH * IMAGE_HEIGHT];
  float* num_spatial_samples_required_in_grid = new float[IMAGE_WIDTH * IMAGE_HEIGHT];

  // Initialize the arrays.
  for (int y = 0; y < IMAGE_HEIGHT; y++) {
    for (int x = 0; x < IMAGE_WIDTH; x++) {
      num_temporal_samples_required_in_grid[y * IMAGE_WIDTH + x] = 0;
      num_spatial_samples_required_in_grid[y * IMAGE_WIDTH + x] = 0;
    }
  }


  vector<ColorSample*>* colors_in_grid;
  colors_in_grid = new vector<ColorSample*>[IMAGE_WIDTH * IMAGE_HEIGHT];

  float density_start = getwalltime();
  fprintf(stderr, "calculate number of samples\n");

  if (set_samples_arg.getValue() == -1 && stratified_x_arg.getValue() == -1) {
      // Calculate the number of samples needed for each pixel.
      for (int y = 0; y < IMAGE_HEIGHT; y++) {
        for (int x = 0; x < IMAGE_WIDTH; x++) {
          if (!IsInRenderBound(x, y)) continue;
          if (segments[y * IMAGE_WIDTH + x].size() > 0) {
            TimeSegmentAtPixel *segment = segments[y * IMAGE_WIDTH + x][0];

            // final 3D sample density, end of sec 6.3
            float timeSamples = NumTemporalSamplesNeeded(segment);
            float motionSamples = NumSpatialSamplesNeeded(segment);
            // packing replicas as close as possible
            float perpSamples = segment->omega_x_max +
                fmin(segment->omega_x_max, OMEGA_PIX);
            float spatialSamples = sqrt(motionSamples * perpSamples);

            FillSamplingRateToNearbyPixels(*segment,
                timeSamples * temporal_sampling_rate_factor_arg.getValue(),
                spatialSamples * spatial_sampling_rate_factor_arg.getValue(),
                num_temporal_samples_required_in_grid,
                num_spatial_samples_required_in_grid);
          }
        }
      }
  }
  float density_end = getwalltime();
  float density_sec = (density_end - density_start);

  fprintf(stderr, "write out visualization\n");
  // Intermediate data visualization.
  if (1 || output_sampling_density_map_arg.getValue() ||
      output_filter_width_map_arg.getValue()) {
    int total_samples_needed = 0;

    float* img2 = new float[IMAGE_WIDTH * IMAGE_HEIGHT * 4];
    memset(img2, 0, IMAGE_WIDTH * IMAGE_HEIGHT * 4 * sizeof(float));

    for (int py = 0; py < IMAGE_HEIGHT; py++) {
      for (int px = 0; px < IMAGE_WIDTH; px++) {
        float *pixel_color = img + (py * IMAGE_WIDTH + px) * 4;
        float *pixel_color2 = img2 + (py * IMAGE_WIDTH + px) * 4;
        pixel_color[0] = 0;
        pixel_color[1] = 0;
        pixel_color[2] = 0;
        pixel_color[3] = 1;

        pixel_color2[0] = 0;
        pixel_color2[1] = 0;
        pixel_color2[2] = 0;
        pixel_color2[3] = 1;

        if (!IsInRenderBound(px, py))
          continue;

        if (segments[py * IMAGE_WIDTH + px].size() == 0)
          continue;

        // Show the number of samples needed
        if (1 || output_sampling_density_map_arg.getValue()) {
          pixel_color[0] = num_temporal_samples_required_in_grid[py * IMAGE_WIDTH + px]
                         * num_spatial_samples_required_in_grid[py * IMAGE_WIDTH + px];
          pixel_color[1] = num_temporal_samples_required_in_grid[py * IMAGE_WIDTH + px];
          pixel_color[2] = num_spatial_samples_required_in_grid[py * IMAGE_WIDTH + px];
          // pixel_color[3] = segments[py * IMAGE_WIDTH + px][0]->wide_filter_radius;

          pixel_color2[0] = segments[py * IMAGE_WIDTH + px][0]->a_min;
          pixel_color2[1] = segments[py * IMAGE_WIDTH + px][0]->a_max;
          pixel_color2[2] = segments[py * IMAGE_WIDTH + px][0]->omega_x_max;
          pixel_color2[3] = segments[py * IMAGE_WIDTH + px][0]->refined_filter_radius;
        }

        if (output_filter_width_map_arg.getValue()) {
          pixel_color[0] = GetOptimalFilterWidth(
              segments[py * IMAGE_WIDTH + px][0]->a_min,
              segments[py * IMAGE_WIDTH + px][0]->a_max,
              segments[py * IMAGE_WIDTH + px][0]->omega_x_max);
        }
      }
    }
    char density_filename[1024];
    sprintf(density_filename, "./%s_%.3f_density.exr", RIB_FILENAME, animation_start_arg.getValue());
    WriteExr(density_filename, img, IMAGE_WIDTH, IMAGE_HEIGHT);
    memset(img, 0, IMAGE_WIDTH * IMAGE_HEIGHT * 4 * sizeof(float));

    sprintf(density_filename, "./%s_%.3f_stats.exr", RIB_FILENAME, animation_start_arg.getValue());
    WriteExr(density_filename, img2, IMAGE_WIDTH, IMAGE_HEIGHT);
    memset(img2, 0, IMAGE_WIDTH * IMAGE_HEIGHT * 4 * sizeof(float));
  }

  vector<InputSample> moreInSamples;
  vector<ColorSample> moreOutSamples;

#ifdef QUAD_SCENE_VERIFICATION
  // HACK: only for the quad scene. Uniform sampling rate across the image.
  GenerateSamples(IMAGE_WIDTH, IMAGE_HEIGHT,
      TIME_START, TIME_END,
      (int) ceil(IMAGE_WIDTH * num_spatial_samples_required_in_grid[256][256]),
      IMAGE_HEIGHT,
      num_temporal_samples_required_in_grid[256][256] * (TIME_END - TIME_START),
      RENDER_LEFT, RENDER_RIGHT,
      RENDER_TOP, RENDER_BOTTOM,
      STRATIFIED_SAMPLING, moreInSamples);

#else
  fprintf(stderr, "generate sample locations\n");
if (stratified_x_arg.getValue() > -1) {
    // stratified sampling (just like brute force)
    GenerateSamples(IMAGE_WIDTH, IMAGE_HEIGHT,
                    TIME_START, TIME_END,
                    IMAGE_WIDTH * stratified_x_arg.getValue(),
                    IMAGE_HEIGHT * stratified_x_arg.getValue(),
                    (int)(stratified_t_arg.getValue() * (TIME_END - TIME_START)),
                    0, IMAGE_WIDTH,
                    0, IMAGE_HEIGHT,
                    true, moreInSamples);

} else {
  // For general scene
  const int BLOCK_SIZE = 50;
  int numXBlocks = (IMAGE_WIDTH + BLOCK_SIZE - 1) / BLOCK_SIZE;
  int numYBlocks = (IMAGE_HEIGHT + BLOCK_SIZE - 1) / BLOCK_SIZE;
  for (int yblock = 0; yblock < numYBlocks; yblock++) {
    for (int xblock = 0; xblock < numXBlocks; xblock++) {
      int yblockend = MIN(IMAGE_HEIGHT, (yblock + 1) * BLOCK_SIZE);
      int xblockend = MIN(IMAGE_WIDTH, (xblock + 1) * BLOCK_SIZE);
      for (int y = yblock * BLOCK_SIZE; y < yblockend; y++) {
        for (int x = xblock * BLOCK_SIZE; x < xblockend; x++) {

          if (!IsInRenderBound(x, y)) continue;
          int spatial_sampling_rate;
          // Only support integer sampling rate.
          spatial_sampling_rate = (int) (0.5 + num_spatial_samples_required_in_grid[y * IMAGE_WIDTH + x]);
          spatial_sampling_rate = MAX(1, spatial_sampling_rate);
          int num_temporal_samples =
              (int) (0.5 + num_temporal_samples_required_in_grid[y * IMAGE_WIDTH + x] *
                         (TIME_END - TIME_START));
          num_temporal_samples = MAX(1, num_temporal_samples);
          

          // we always have at least one sample per pixel per shutter
          float spacerate = MAX(1.0, num_spatial_samples_required_in_grid[y * IMAGE_WIDTH + x]);
          float timerate = MAX(1.0, num_temporal_samples_required_in_grid[y * IMAGE_WIDTH + x]);
          int totalSamples = (int) ceil(spacerate * spacerate * timerate);
          totalSamples = MAX(1, totalSamples);

          int uniform_samples = spatial_sampling_rate * spatial_sampling_rate * num_temporal_samples;
          int maxSamples = (int) max_samples_arg.getValue();

          
          if (set_samples_arg.getValue() > 0) {
              totalSamples = set_samples_arg.getValue();
          }

          if (!uniform_arg.getValue() || uniform_samples > maxSamples) {
              //  x, y, spacerate, timerate, totalSamples);
              HaltonSampler sampler;
              totalSamples = MIN(maxSamples, totalSamples);
#if 0
              // this actually hurts, and causes more discontinuities
              if (totalSamples < 8) {
                sampler.CenterSequence(totalSamples);
              }
#endif
              if (totalSamples > maxSamples / 2) {
                sampler.Randomize();
              }

              // reflect sequence for odd pixels
              // (sequence is centered so that the first sample is always in
              // the center)
              if ((x + y) % 2 == 0) {
                  for (int i = 0; i < totalSamples; i++) {
                        InputSample curSample = sampler.GetNextSample();
                        curSample.x = x + curSample.x;
                        curSample.y = y + curSample.y;
                        moreInSamples.push_back(curSample);
                  }
              } else {
                  // reflect sequence for odd pixels
                  for (int i = 0; i < totalSamples; i++) {
                        InputSample curSample = sampler.GetNextSample();
                        curSample.x = x + 1.0 - curSample.x;
                        curSample.y = y + 1.0 - curSample.y;
                        curSample.time = 1.0 - curSample.time;
                        moreInSamples.push_back(curSample);
                  }
              }
          } else {
              // uniform sampling, limit spatial and temporal sampling
              if (spatial_sampling_rate > 1 && num_temporal_samples > 16) {
                spatial_sampling_rate = MIN(2, spatial_sampling_rate);
                num_temporal_samples = MIN(64, num_temporal_samples);
              } else if (num_temporal_samples > 16) {
                num_temporal_samples = MIN(128, num_temporal_samples);
              } else {
                spatial_sampling_rate = MIN(4, spatial_sampling_rate);
                num_temporal_samples = MIN(64, num_temporal_samples);
              }
              for (int super_x = 0; super_x < spatial_sampling_rate; super_x++) {
                for (int super_y = 0; super_y < spatial_sampling_rate; super_y++) {
                  for (int ti = 0; ti < num_temporal_samples; ti++) {
                    InputSample curSample;
                    curSample.x = ((float) x + (super_x + GetOffset()) / spatial_sampling_rate);
                    curSample.y = ((float) y + (super_y + GetOffset()) / spatial_sampling_rate);
                    curSample.time = ((float) ti + GetOffset()) / num_temporal_samples
                        * (TIME_END - TIME_START) + TIME_START;
                    moreInSamples.push_back(curSample);
                  }
                }
              }
          } // uniform
        } // xpix
      } // ypix
    } // xblock
  } // yblock
} // if stratified
#endif

  fprintf(stderr, "render new sample locations\n");
  // Render new samples
  float final_sample_start = getwalltime();
  if (moreInSamples.size()) {
    if (animation_start_arg.getValue() > 0 && animation_length_arg.getValue() > 0) {
      RenderColorSamplesMultiFile(moreInSamples, moreOutSamples,
          IMAGE_WIDTH, IMAGE_HEIGHT, RIB_FILENAME,
          PASS1_SAMPLE_INPUT_FILENAME, PASS1_SAMPLE_RESULTS_FILENAME,
          !norerun_pixie_pass2_arg.getValue(), keep_sample_files_arg.getValue(),
          animation_start_arg.getValue(), 1.0 / animation_length_arg.getValue());
    } else {
      RenderColorSamples(moreInSamples, moreOutSamples, IMAGE_WIDTH, IMAGE_HEIGHT,
          RIB_FILENAME,
          PASS2_SAMPLE_INPUT_FILENAME, PASS2_SAMPLE_RESULTS_FILENAME,
          !norerun_pixie_pass2_arg.getValue(), keep_sample_files_arg.getValue());
    }
  }
  float final_sample_end = getwalltime();
  float final_sample_sec = (final_sample_end - final_sample_start);

  fprintf(stderr, "process new sample data\n");
  // Put new out samples in the grid
  for (int i = 0; i < moreOutSamples.size(); i++) {
    ColorSample curSample = moreOutSamples[i];
    int xpix = (int) (curSample.x);
    int ypix = (int) (curSample.y);
    if (xpix < 0 || xpix >= IMAGE_WIDTH || ypix < 0 || ypix >= IMAGE_HEIGHT) {
      cout << "ERROR: out of image bound. " << curSample.x << ", " << curSample.y << endl;
      continue;
    }
    colors_in_grid[ypix * IMAGE_WIDTH + xpix].push_back(&moreOutSamples[i]);
  }

  // Store the sampling density information in each sample.
  for (int y = 0; y < IMAGE_HEIGHT; y++) {
    for (int x = 0; x < IMAGE_WIDTH; x++) {
      for (int i = 0; i < colors_in_grid[y * IMAGE_WIDTH + x].size(); i++) {
        colors_in_grid[y * IMAGE_WIDTH + x][i]->density =
            colors_in_grid[y * IMAGE_WIDTH + x].size();
      }
      img[(y * IMAGE_WIDTH + x) * 4 + 0] = colors_in_grid[y * IMAGE_WIDTH + x].size();
    }
  }
  char samples_filename[1024];
  sprintf(samples_filename, "./%s_%.3f_samples.exr", RIB_FILENAME, animation_start_arg.getValue());
  WriteExr(samples_filename, img, IMAGE_WIDTH, IMAGE_HEIGHT);
  memset(img, 0, IMAGE_WIDTH * IMAGE_HEIGHT * 4 * sizeof(float));

  float reconstruct_start = getwalltime();
  fprintf(stderr, "reconstruct final image\n");
  int warning_count = 0;
  // Image reconstruction
  for (int py = 0; py < IMAGE_HEIGHT; py++) {
    for (int px = 0; px < IMAGE_WIDTH; px++) {
      if (!IsInRenderBound(px, py)) continue;
      float *pixel_color = img + (py * IMAGE_WIDTH + px) * 4;
      for (int segi = 0; segi < segments[py * IMAGE_WIDTH + px].size(); segi++) {
        TimeSegmentAtPixel *segment = segments[py * IMAGE_WIDTH + px][segi];
        float weight_sum = 0;
        float segment_color[] = {0, 0, 0, 0};
        vector<pair<ColorSample*, float> > samples;
        // Find the samples inside the filter and compute their weights.
        ApplyFilter(*segment, filter_func, colors_in_grid, &samples,
            nobucket_arg.getValue(), oldbucket_arg.getValue());
        for (int i = 0; i < samples.size(); i++) {
          weight_sum += samples[i].second;
          multadd(segment_color, 1, samples[i].first->color, samples[i].second, segment_color);
        }
        if (weight_sum < 0.000001) {
            if (warning_count % 1000 == 0) {
               cout << "warning " << warning_count << ", " << px << ", " << py << " weight too low: " << weight_sum << ", " << samples.size() << ", no color." << endl;
            }
          warning_count++;
        } else {
          multadd(pixel_color, 1, segment_color, segment->time_interval / weight_sum, pixel_color);
        }
        pixel_color[3] = 1;
      }
    }

    // Output a temp image when finishing each scanline.
    if (py < RENDER_TOP || py > RENDER_BOTTOM) continue;
    char filename[1024];
    sprintf(filename, "./partial%03d.exr", py);
    //WriteExr(filename, img, IMAGE_WIDTH, IMAGE_HEIGHT);
  }

  fprintf(stderr, "clear datastructures\n");
  for (int i = 0; i < IMAGE_WIDTH * IMAGE_HEIGHT; i++) {
    for (int j = 0; j < segments[i].size(); j++) {
        delete segments[i][j];
    }
  }
  float reconstruct_end = getwalltime();
  float reconstruct_sec = (reconstruct_end - reconstruct_start);

  moreOutSamples.clear();
  delete[] colors_in_grid;
  delete[] segments;
  delete[] num_temporal_samples_required_in_grid;
  delete[] num_spatial_samples_required_in_grid;

  fprintf(stderr, "write final image\n");
  // Output the final image.
  WriteExr(output_filename_arg.getValue().c_str(), img, IMAGE_WIDTH, IMAGE_HEIGHT);
  delete[] img;

  float program_end = getwalltime();
  float program_sec = (program_end - program_start);


  float sample_sum_sec = initial_sample_sec + final_sample_sec;
  float misc_sec = program_sec - (initial_sample_sec + final_sample_sec + refine_sec + density_sec + reconstruct_sec); 
  float us_sum_sec = refine_sec + density_sec + reconstruct_sec + misc_sec;
  int program_min = (int) (program_sec / 60);

  fprintf(stderr, "\n\n"
    "RENDERING                                  \n"
    "  initial samples:    %8.1f sec  (% 5.1f%%)\n"
    "  final samples:      %8.1f sec  (% 5.1f%%)\n"
    "  -----------------------------------------\n"
    "                      %8.1f sec  (% 5.1f%%)\n"
    "                                           \n"
    "OUR METHOD                                 \n"
    "  refinement:         %8.1f sec  (% 5.1f%%)\n"
    "  sample densities:   %8.1f sec  (% 5.1f%%)\n"
    "  reconstruction:     %8.1f sec  (% 5.1f%%)\n"
    "  misc:               %8.1f sec  (% 5.1f%%)\n"
    "  -----------------------------------------\n"
    "                      %8.1f sec  (% 5.1f%%)\n"
    "                                           \n"
    "                                           \n"
    "TOTAL                 %8.1f sec,    %d min %2d sec\n",
    initial_sample_sec, 100 * initial_sample_sec / program_sec,
    final_sample_sec, 100 * final_sample_sec / program_sec,
    sample_sum_sec, 100 * sample_sum_sec / program_sec,
    refine_sec, 100 * refine_sec / program_sec,
    density_sec, 100 * density_sec / program_sec,
    reconstruct_sec, 100 * reconstruct_sec / program_sec,
    misc_sec, 100 * misc_sec / program_sec,
    us_sum_sec, 100 * us_sum_sec / program_sec,
    program_sec, program_min, (int) (program_sec - (60 * program_min)));

    return 0;
}

void ConstructSegments(vector<FreqSample*>* samples_in_grid,
                       vector<TimeSegmentAtPixel*>* segments) {

  // Find avg, max and min motion for each pixel.
  for (int y = 0; y < IMAGE_HEIGHT; y++) {
    for (int x = 0; x < IMAGE_WIDTH; x++) {
      if (!IsInRenderBound(x, y)) {
        segments[y * IMAGE_WIDTH + x].push_back(new TimeSegmentAtPixel());
        TimeSegmentAtPixel* segment = segments[y * IMAGE_WIDTH + x][0];
        segment->start_time = 0;
        segment->end_time = SHUTTER_END;
        segment->x = x + 0.5;
        segment->y = y + 0.5;
        segment->avg_motion = FindAvgMotion(segment->samples);
        segment->a_min = FindMinMotionLength(segment->samples);
        segment->a_max = FindMaxMotionLength(segment->samples);
        segment->omega_x_max = FindMaxOmegaXMax(segment->samples);
        continue;
      }
      // TODO Iterate through samples and segment them into groups with similar
      // motions.
      vector<FreqSample*> samples;

      // advect other nearby samples through a gather approach
      // TODO a scatter approach is more difficult, but probably more ideal
      CollectSamples(samples_in_grid, Vector2D(x + 0.5, y + 0.5), &samples);

      sort(samples.begin(), samples.end(), SampleCMP);
      // FIXME: Revise this section if we want to enable segmentation.
      for (int i = 0; i < samples.size(); i++) {
        // Create the first group.
        if (i == 0) {
          segments[y * IMAGE_WIDTH + x].push_back(new TimeSegmentAtPixel());
          segments[y * IMAGE_WIDTH + x].back()->start_time = 0;
        // See a different motion, create a new group.
        } else if (SEGMENTATION &&
              (samples[i - 1]->v2d_motion - samples[i]->v2d_motion).length() > 5) {
          float separate_point = (samples[i - 1]->shift_time + samples[i]->shift_time) / 2;
          segments[y * IMAGE_WIDTH + x].back()->end_time = separate_point;
          TimeSegmentAtPixel *segment = new TimeSegmentAtPixel();
          segment->start_time = separate_point;
          segments[y * IMAGE_WIDTH + x].push_back(segment);
        }
        // Put the sample into the last segment.
        segments[y * IMAGE_WIDTH + x].back()->samples.push_back(samples[i]);
      }
      segments[y * IMAGE_WIDTH + x].back()->end_time = SHUTTER_END;

      // Calculate the avg, min and max velocities for each segment.
      for (int i = 0; i < segments[y * IMAGE_WIDTH + x].size(); i++) {
        TimeSegmentAtPixel* segment = segments[y * IMAGE_WIDTH + x][i];
        segment->time_interval = segment->end_time - segment->start_time;
        segment->x = x + 0.5;
        segment->y = y + 0.5;
        segment->avg_motion = FindAvgMotion(segment->samples);
        segment->a_min = FindMinMotionLength(segment->samples);
        segment->a_max = FindMaxMotionLength(segment->samples);
        segment->omega_x_max = FindMaxOmegaXMax(segment->samples);

        // Check motion direction consistency among advected samples.
        for (int j = 0; j < segment->samples.size(); j++) {
          if (HasSomeFrequencyContent(segment->samples[j]->omega_x_max, segment->samples[j]->a_max) &&
                !IsConsistentDirections(segment->avg_motion, segment->samples[j]->v2d_motion)) {

            segment->a_min = 0;
          }
        }
      }
    }
  }
}

